package httpcats

  1. Overview
  2. Docs

Module type Runtime.CONNECTIONSource

Sourcetype conn
Sourcetype flow
Sourceval run : conn -> ?tags:Logs.Tag.set -> ?read_buffer_size:int -> ?upgrade:(flow -> unit) -> flow -> unit Miou.t

run conn ?read_buffer_size flow runs the given state-machine conn on top of flow.

read_buffer_size is the initial size of the bigstring where we accumulate what the peer sends us (it grows on demand); it should be H1.Config.read_buffer_size / H2.Config.read_buffer_size.