package httpcats

  1. Overview
  2. Docs

Module Httpcats.FlowSource

Flow is the interface required by httpcats to implement the HTTP client and the HTTP server. This interface is really close to what Unix can provide.

Sourceexception Closed_by_peer
Sourcemodule type S = sig ... end