package httpcats

  1. Overview
  2. Docs

v0.2.0 (2026-02-19) Paris - France

  • Improve the OPAM file and the README.md (@geazi-anc, #39)
  • Split out the core of httpcats to be re-usable by unikernels (@dinosaure, #40)
  • Avoid a memory leak about logs (@dinosaure, #42)
  • Upgrade GitHub CI and ocamlformat (@dinosaure, #43)

v0.1.0 (2025-08-09) Paris - France

  • Add the support of websocket (@swrup & @dinosaure)
  • Improve the Httpcats.request and handle cookies and redirections well (@dinosaure)
  • Improve the runtime implement to process an HTTP communication (with http/1.1 & h2) (@dinosaure)
  • Improve the documentation (@dinosaure)

This changelog does not contain PRs referring to additions to httpcats, as they were produced with a quick iteration to finalise the release of httpcats.0.1.0.

v0.0.1 (2024-09-13) Paris - France

  • First release